/**
 * This script handles the retrieval of properties.  It is suitable for usage from other scripts via
 * evaluate(new File(..., 'getAHPProperties.groovy'))
 * 
 * When invoked via the evaluate method the arguments are inherited from the calling script.
 * If the properties argument is supplied then this will simply parse those and return a Map of their values.
 * Otherwise it will contact the AHP server and retrieve all step and job properties.
 * 
 * @param an optional argument of properties formatted according to the java.util.Properties#store and #load formats.
 * @return the property values.
 */

final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
final def ahptool = isWindows ? 'ahptool.cmd' : 'ahptool'
        
/**
 * Get the properties from the AHP server
 * 
 * @param scopes collection of scope names (job, workflow, buildlife, etc) to get the properties for
 */
def props = [:]

if (args) {
    // read a properties argument if present
    def newProps = new Properties()
    newProps.load(new ByteArrayInputStream(args[0].getBytes()))
    props.putAll(newProps)
}
else {
    // otherwise get properties from server
    ['Job', 'Step'].each { scope ->
        def ahptoolProc = [ahptool, "get${scope}Properties"].execute()
        ahptoolProc.consumeProcessErrorStream(System.out);
        def propertiesXml = ahptoolProc.text
        ahptoolProc.waitFor()
        if (ahptoolProc.exitValue()) {
            throw new Exception("Failed to get step properties from ahptool: " + propertiesXml)
        }
        new XmlSlurper().parseText(propertiesXml).property.each{ propElem ->
            props[propElem.@name.text()] = propElem.text()
        }
    }
}

return props;