/**
 *
 * Perform:
 *   P4  Login                   - if desired
 *   P4  Create ClientSpec       - if needed
 *   P4  Changes                 - filtered to ones matching source and exclusion/filters
 *   AHP Upload Changelog Report
 *   AHP Upload Changes to DB    - if desired
 *   P4  Delete ClientSpec       - if created
 *   P4  Logout                  - if login was performed
 *
 */
import java.text.SimpleDateFormat

final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
final def ahptool = isWindows ? 'ahptool.cmd' : 'ahptool'

private final def GMT = TimeZone.getTimeZone("GMT")

private final def AHP_DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.S Z");
AHP_DATE_FORMAT.timeZone = GMT

private final def PERFORCE_CHANGELOG_OUT_DATE = new java.text.SimpleDateFormat("yyyy/MM/dd HH:mm:ss")
private final def CHANGE_SUMMARY_PATTERN = ~'(?m)^Change \\d+ by (\\S+)@\\S+ on (\\d+/\\d+/\\d+ \\d+:\\d+:\\d+)'

/**
 *
 */
public String formatChangelogDateParameter(Long commandDate, Double timeDifference){
    if (!commandDate) {
        return null
    }

    // adjust date for difference between ahp-agent and p4-server
    if (timeDifference) {
        def p4Cal = Calendar.instance
        p4Cal.time = new Date(commandDate)
        p4Cal.add(Calendar.HOUR_OF_DAY, (int)timeDifference)
        int minDiff = (int)(timeDifference * 60) % 60
        if (minDiff) {
            p4Cal.add(Calendar.MINUTE, minDiff)
        }
        commandDate = p4Cal.time.time
    }

    def PERFORCE_CHANGELOG_DATE = new java.text.SimpleDateFormat("yyyy/MM/dd:HH:mm:ss")
    return PERFORCE_CHANGELOG_DATE.format(new Date(commandDate))
}

/**
 * Converts an Ant-style wild-mat string into an equivalent java.util.regex.Pattern
 *   ? is a single char wildcard
 *   * is a wildcard matches any string (but does not cross directories)
 *   ** is a wildcard which matches any string and can cross any number of directories
 */
public def convertToPattern( def antWildMat) {
    // trim leading / character from pattern
    def pattern = antWildMat.startsWith('/') ? antWildMat : '/'+antWildMat

    // deal with special regex-characters that should be interpreted as literals
    '\\.+[]^${}|()'.toCharArray().each{ c ->
        pattern = pattern.replace(''+c, '\\'+c)
    }
    pattern = pattern.replace('?', '.') // ? is a single-char wildcard

    // deal with ant-style wildcards
    StringBuffer result = new StringBuffer()
    def m = (pattern =~ '\\*\\*/|\\*\\*|\\*')
    while (m) {
        def token = m.group()
        def replacement;
        if (token == '**/') {
            replacement = '.*(?<=/)'
        }
        else if (token == '**') {
            replacement = '.*'
        }
        else {
            replacement = '[^/]*'
        }
        m.appendReplacement(result, java.util.regex.Matcher.quoteReplacement(replacement))
    }
    m.appendTail(result)
    return ~(result.toString())
}


//------------------------------------------------------------------------------
// GET ALL INPUT PARAMETERS and VALIDATE STATE
//------------------------------------------------------------------------------

final def  workDir                   = new File('.').canonicalFile
final def  out                       = System.out
final File resourceHome              = new File(System.getenv()['PLUGIN_HOME'])
final def  props                     = evaluate(new File(resourceHome, 'getAHPProperties.groovy'))

final def client                     = props['source/client']
final boolean isLoggingIn            = props['source/isLoggingIn']?.toBoolean() // TODO should be a source/repo prop?
final def commandPath                = props['source/repo/commandPath'] ?: 'p4'
final def port                       = props['source/repo/port']
final def username                   = props['source/repo/username']
final def password                   = props['source/repo/password'] ?: props['source/repo/passwordScript']
final def charset                    = props['source/repo/charset']
final def commandCharset             = props['source/repo/commandCharset']

final def userExcludes       = props['source/userExcludeArray']?.readLines() ?: []
final def fileFilters        = props['source/filepathExcludeArray']?.readLines() ?: []

final Double timeDifference  = props['timezoneDifference']?.toDouble()
def ticketFile               = props['ticketFile'] != null ? new File(props['ticketFile']) : null;

final def startDateString    = formatChangelogDateParameter(props['startDate']?.toLong(), timeDifference)
final def endDateString      = formatChangelogDateParameter(props['endDate']?.toLong(), timeDifference)
final def persisteToDB       = props['persistingChanges']?.toBoolean()

if (userExcludes) {
  userExcludes = userExclude.collect{it.trim()}.findAll{it}
}

def includeFiles = []
def excludeFiles = []
if (fileFilters) {
  fileFilters.collect{it.trim()}.findAll{it}.each{filter ->
      def pattern = convertToPattern(filter.substring(1))
      if (filter.startsWith('+')) {
          includeFiles << pattern
      }
      else if (filter.startsWith('-')) {
          excludeFiles << pattern
      }
      else {
          // not a valid expression
      }
  }
}

//create a ProcessBuilder with appropriate ticket file in the environment and the working directory
def pb = new ProcessBuilder(['echo'] as String[]).directory(workDir)


//------------------------------------------------------------------------------
// Define convenience methods which depend upon input parameters
//------------------------------------------------------------------------------

/**
 * A convenience method for running commands and optionally parsing the stdout of that command.
 * If closure is non-null, the closure will be passed the resultant {@link Process} and is expected to deal with all IO.
 * Otherwise, the process' stdOut and stdErr are forwarded to this scripts 'stdOut' and stdIn is untouched.
 *
 * @param message an optional message to print prior to the commandline
 * @param command the command to be used as a String[]
 * @param closure an optional closure to deal with Process IO
 */
def runCommand = {def message, def command, def closure ->
    pb.command(command as String[])
    println()
    if (message) {
        println(message)
    }
    println("command: ${pb.command().join(' ')}")
    def proc = pb.start()
    if (closure) {
        closure(proc)
    }
    else {
        proc.out.close() // close stdin
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
    proc.waitFor()
    if (proc.exitValue()) {
       throw new Exception("Command failed with exit code: "+proc.exitValue())
    }
}

/**
 * Create a the list of initial arguments needed for perforce commands
 */
def createP4BaseCommand = {
    def command =  [commandPath, '-c', client, '-p', port, '-u', username]
    if (password  && !ticketFile?.exists()) {
        command << '-P' << password
    }
    if (charset) {
        command << '-C' << charset
    }
    if (commandCharset) {
        command << '-Q' << commandCharset
    }
    return command
}

/**
 * determine if the given path is a non-excluded path belonging to the desired set
 *
 * @param path the file path to be tested
 */
def pathIsIncluded = { path ->
    boolean included = true;

    // #1 Check belongs to a vaid depot?
    //     TODO

    // #2 Calculate relative path and check against inclue/exclude patterns

    def relativePath = '' // TODO

    for (include in includeFiles) {
        if (relativePath ==~ include) {
            return true
        }
    }

    for (exclude in excludeFiles) {
        if (relativePath ==~ exclude) {
            return false
        }
    }
    return true;
}

//------------------------------------------------------------------------------
// P4 Login
//------------------------------------------------------------------------------
if (ticketFile?.exists()) {
    pb.environment()['P4TICKETS'] = ticketFile.absolutePath
}
else if (isLoggingIn) {
    ticketFile = File.createTempFile('ah3agent', 'ticket').canonicalFile
    props['ticketFile'] = ticketFile.absolutePath

    pb.environment()['P4TICKETS'] = ticketFile.absolutePath

    def loginCmd = createP4BaseCommand()
    loginCmd << "login"
    runCommand('P4 Login', loginCmd) { proc ->
        if (password) {
            proc.out << password
        }
        proc.out.close()
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
}

//------------------------------------------------------------------------------
// P4 Create ClientSpec
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
// get changesets IDs of interest
//------------------------------------------------------------------------------

def changes = ''

def command = createP4BaseCommand()
command << 'changes' << '-s' << 'submitted'
if (startDateString) {
    command << "//${client}/...@${startDateString},${endDateString}"
}
else if (endDateString) { // isn't this always true?
    command << "//${client}/...@${endDateString}"
}
else {
    command << "//${client}/..."
}

runCommand('Getting Changes', command) { proc ->
  proc.out.close()                    // close   the stdIn  of the process
  proc.consumeProcessErrorStream(System.out) // forward the stdErr of the process
  changes = proc.text                 // capture the stdout of the process
  println changes                     // echo captured stdout
}
println("found ${changes.readLines().size()} change-sets")

//------------------------------------------------------------------------------
// Process output into changesets
//------------------------------------------------------------------------------


def typeMap = ['add':'A', 'edit':'M', 'delete':'D']

def changeSets = []

// Change X on yyyy/MM/dd by user@client 'comment'
changes.eachMatch(/(?m)^Change (\d+)/){ changematch, changeId ->
    def changeSet = [id:changeId, user:null, message:null, date:null, fileSet:[]]

    // Get details
    def summary = ''
    def descCommand = createP4BaseCommand()
    descCommand << 'describe' << '-s' << changeId
    runCommand(null, descCommand) { proc ->
        proc.out.close()                    // close   the stdIn  of the process
        proc.consumeProcessErrorStream(System.out) // forward the stdErr of the process
        summary = proc.text                 // capture the stdout of the process
        println summary                     // echo captured stdOut
    }

    // Summary Format:
    //   Change X by user@client on yyyy/MM/dd HH:mm:ss
    //   <some number of lines w/ or w/o content>
    //   Affected files ...
    //   ... //depot/file#3 edit
    //   ... //depot/file#3 edit
    //   ... //depot/file#3 edit
    //   ... //depot/file#3 edit
    //

    // first line is "Change X by user@client on yyyy/MM/dd HH:mm:ss"
    summary.find(CHANGE_SUMMARY_PATTERN) { summarymatch, user, dateString ->
        def p4Date = PERFORCE_CHANGELOG_OUT_DATE.parse(dateString) // TODO deal with 'unknown' date

        // Pick out the changelog comment (anything between the first line and "Affected files")
        int commentStart = summary.indexOf(summarymatch)+summarymatch.length()
        int commentEnd = summary.indexOf("Affected files ...", commentStart)
        if (commentStart > -1 && commentEnd > -1 && commentEnd > commentStart) {
            changeSet.message = summary.substring(commentStart, commentEnd).trim()
        }

        changeSet.user = user
        changeSet.date = AHP_DATE_FORMAT.format(p4Date)

        boolean hasAllowedAuthor = !userExcludes.contains(changeSet.user)
        boolean hasAllowedFile = false

        // Evenatially there is a line "Affected files ..."
        //  followed by lines of "... //depot/file#3 edit"
        summary.eachMatch(/(?m)^\.\.\. (.+)#\d+ (\S+)/){affectedmatch, file, operation ->
            def changeType = typeMap[operation] ?: 'M' // convert operation to A, M, D types
            changeSet.fileSet << [type: changeType, rev:null, path:file]
            hasAllowedFile =  hasAllowedFile || pathIsIncluded(file)
        }

        if (hasAllowedAuthor && hasAllowedFile) {
            changeSets << changeSet
        }
    }
}

//------------------------------------------------------------------------------
// Upload a changelog report (non-database)
//------------------------------------------------------------------------------

// TODO create and upload a report of the changes?

//------------------------------------------------------------------------------
// Construct XML message and upload to AHP Database
//------------------------------------------------------------------------------

if (!changeSets) {
    println "No changes detected"
}
else if (persisteToDB) {
    def xml = groovy.xml.DOMBuilder.newInstance().'change-log'{
        changeSets.each{ changeSet ->
            'change-set'(){
                id(changeSet.id)
                user(changeSet.user)
                date(changeSet.date)
                if (changeSet.message) {
                    comment(changeSet.message)
                }
                //module('')
                //branch('')
                'repository-type'('perforce')
                'repository-id'(System.getenv()['AH_REPOSITORY_ID'])
                'file-set'{
                    changeSet.fileSet.each{change ->
                        file('change-type':change.type, 'revision-number':change.rev, change.path)
                    }
                }
            }
        }
    }

    runCommand('Uploading ChangeSets to AHP', [ahptool, 'addSourceChanges', '-']) { proc ->
        proc.consumeProcessOutput(out, out) // forward process stdout and stderr to our stdout
        proc.withWriter{ it << xml }
    }
}
else {
    println "Not persisting changes to database"
}

//------------------------------------------------------------------------------
//P4 Logout
//------------------------------------------------------------------------------
if (isLoggingIn && ticketFile?.exists()) {
   def logoutCmd = createP4BaseCommand()
   logoutCmd << "logout"
   runCommand('P4 Logout', logoutCmd, null)
}