import java.util.regex.Matcher

/**
 * Handle all substeps involed in performing a p4 clientspec creation
 *
 * 1) P4  login             - if desired
 * 2) P4  get clientspec template - if desired
 * 3) P4  create clientspec - if using template, also edit spec if using template with custom work-dir
 * 4) P4  logout            - if login was performed
 *
 * @param a properties-formated argument containing all the need data
 */

final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
final def ahptool = isWindows ? 'ahptool.cmd' : 'ahptool'

/**
 * replace path in a configspec with given path
 */
def modifyClientSpecConfigRoot = {def config, def newRoot ->
    if (!config) {
        return config
    }
    else if (!newRoot) {
       return config.replaceAll(/(?mi)Root:.*$/, Matcher.quoteReplacement("Root: ${new File('.').canonicalPath}"))
    }
    return config.replaceAll(/(?mi)Root:.*$/, Matcher.quoteReplacement("Root: ${new File('.', newRoot).canonicalPath}"))
}

//------------------------------------------------------------------------------
// GET ALL INPUT PARAMETERS and VALIDATE STATE
//------------------------------------------------------------------------------

final def  workDir                   = new File('.').canonicalFile
final def  out                       = System.out
final File resourceHome              = new File(System.getenv()['PLUGIN_HOME'])
final def  props                     = evaluate(new File(resourceHome, 'getAHPProperties.groovy'))

println("properties: ${props}")

final def client                     = props['source/client']
final boolean isLoggingIn            = props['source/isLoggingIn']?.toBoolean() // TODO should be a source/repo prop?
final def commandPath                = props['source/repo/commandPath'] ?: 'p4'
final def port                       = props['source/repo/port']
final def username                   = props['source/repo/username']
final def password                   = props['source/repo/password'] ?: props['source/repo/passwordScript']
final def charset                    = props['source/repo/charset']
final def commandCharset             = props['source/repo/commandCharset']

final boolean isUsingTemplate        = props['source/clientSpecPolicy'] == 'template'
final boolean createNewClientSpec    = props['source/clientSpecPolicy'] == 'literal'
final def templateName               = props['source/clientspecTemplateName'] // only if 'source/isUsingTemplate' is true
final def directoryOffset            = props['source/directoryOffset']
final def clientspec                 = props['source/clientspec']

def ticketFile                       = props['ticketFile'] != null ? new File(props['ticketFile']) : null;


//create a ProcessBuilder with appropriate ticket file in the environment and the working directory
def pb = new ProcessBuilder(['echo'] as String[]).directory(workDir)
if (ticketFile?.exists()) {
   pb.environment()['P4TICKETS'] = ticketFile.absolutePath
}

//------------------------------------------------------------------------------
// Stop if not using template
//------------------------------------------------------------------------------

if (!createNewClientSpec && !isUsingTemplate) {
    println('Using neither a template nor creating a clientspec, skipping creation of client-spec')
    return // skip rest of script
}

//------------------------------------------------------------------------------
// Define convenience methods which depend upon input parameters
//------------------------------------------------------------------------------

/**
 * A convenience method for running commands and optionally parsing the stdout of that command.
 * If closure is non-null, the closure will be passed the resultant {@link Process} and is expected to deal with all IO.
 * Otherwise, the process' stdOut and stdErr are forwarded to this scripts 'stdOut' and stdIn is untouched.
 *
 * @param message an optional message to print prior to the commandline
 * @param command the command to be used as a String[]
 * @param closure an optional closure to deal with Process IO
 */
def runCommand = {def message, def command, def closure ->
    pb.command(command as String[])
    println()
    if (message) {
        println(message)
    }
    println("command: ${pb.command().join(' ')}")
    def proc = pb.start()
    if (closure) {
        closure(proc)
    }
    else {
        proc.out.close() // close stdin
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
    proc.waitFor()
    if (proc.exitValue()) {
       throw new Exception("Command failed with exit code: "+proc.exitValue())
    }
}

/**
 * Create a the list of initial arguments needed for perforce commands
 */
def createP4BaseCommand = {
    def command =  [commandPath, '-c', client, '-p', port, '-u', username]
    if (password && !ticketFile?.exists()) {
        command << '-P' << password
    }
    if (charset) {
        command << '-C' << charset
    }
    if (commandCharset) {
        command << '-Q' << commandCharset
    }
    return command
}

/**
 * Create a the list of initial arguments needed for perforce commands
 */
def createClientlessP4BaseCommand = {
    def command =  [commandPath, '-p', port, '-u', username]
    if (password && !ticketFile?.exists()) {
        command << '-P' << password
    }
    if (charset) {
        command << '-C' << charset
    }
    if (commandCharset) {
        command << '-Q' << commandCharset
    }
    return command
}

//------------------------------------------------------------------------------
// P4 Login
//------------------------------------------------------------------------------
boolean doLogout = false
if (ticketFile?.exists()) {
   pb.environment()['P4TICKETS'] = ticketFile.absolutePath
}
else if (isLoggingIn) {
   ticketFile = File.createTempFile('ah3agent', 'ticket').canonicalFile
   props['ticketFile'] = ticketFile.absolutePath

   pb.environment()['P4TICKETS'] = ticketFile.absolutePath

   def loginCmd = createP4BaseCommand()
   loginCmd << "login"
    runCommand('P4 Login', loginCmd) { proc ->
        if (password) {
            proc.out << password
        }
        proc.out.close()
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
   doLogout = true
}

//------------------------------------------------------------------------------
// P4 get clientspec template
//------------------------------------------------------------------------------

if (isUsingTemplate) {
    def command = createClientlessP4BaseCommand()
    command << 'client' << '-o' << '-t' << templateName << client
    runCommand(null, command) { proc ->
        proc.out.close() // close stdin
        proc.consumeProcessErrorStream(System.out)
        clientspec = proc.text
    }
}

//------------------------------------------------------------------------------
// update the clientspec to use our custom directoryOffset
//------------------------------------------------------------------------------

// modify the root in the clientspec config to reflect the directoryOffset if
// one was specified in the command
clientspec = modifyClientSpecConfigRoot(clientspec, directoryOffset)

//------------------------------------------------------------------------------
// P4 Create the clientspec
//------------------------------------------------------------------------------

// run the process
def createClientCommand = createClientlessP4BaseCommand()
createClientCommand << 'client' << '-i'
runCommand('Create Clientspec', createClientCommand) { proc ->
    proc.consumeProcessOutput(out, out)
    proc.withWriter{ it << clientspec} // pipe clientspec to child process stdIn and close the stream afterwards
}

//------------------------------------------------------------------------------
// P4 Logout
//------------------------------------------------------------------------------
if (doLogout && ticketFile?.exists()) {
   def logoutCmd = createP4BaseCommand()
   logoutCmd << "logout"
   runCommand('P4 Logout', logoutCmd, null)
}