/**
 * 1) P4  login           - if desired
 * 2) P4  get clientspec
 * 3) P4  get server info
 * 4) AHP set properties
 * 5) P4  logout          - if login was performed
 * 
 */
 
final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
final def ahptool = isWindows ? 'ahptool.cmd' : 'ahptool'

final def ROOT_TOKEN        ='Root:'
final def VIEW_TOKEN        ='View:'
final def SERVER_DATE_TOKEN ='Server date:'
final def PERFORCE_DATE     = new java.text.SimpleDateFormat("yyyy/MM/dd HH:mm:ss")

/**
 * @param base the base time in milliseconds since Epoc
 * @param target the target time in milliseconds since Epoc
 * @return the number of hours between the two times (rounded to nearest half-hour)
 */
public double getHourDifference(long base, long target) {
    double difference = (double)(base - target) / (1000*60*60) // difference in hours
    return Math.round(difference / 0.5) * 0.5                  // round to nearest 1/2 hour
}

/**
 *  read stdout of get-client-info afor the workspacepath and depoPathArray
 *  @param br the buffered reader of the p4 command output
 *  @return a map containing 'rootPath':String and 'depoPaths':String[]
 */
def readClientInfo = {def br ->
    def result = [rootPath: null, depoPaths: []]

    String line = null
    while( (line = br.readLine()) != null )  {
        if (line.startsWith(ROOT_TOKEN)) {
            result.rootPath = line.substring(ROOT_TOKEN.length()).trim()
        }
        else if (line.startsWith(VIEW_TOKEN)) {
            while (line = br.readLine()?.trim()) { // read until we hit a blank line (or null)
                // match first token '//....' where ... has no white-space OR '"//...."' where ... has no quote
                def m = ( line =~ '((?:"//[^"]+)|(?://\\S+))' )
                m.find()
                result.depoPaths << m.group(1)
            }
        }
    }

    return result
}

/**
 * read the timezone offset between client and server
 * @param the server info output
 * @return the number of hourse difference between local time and the server time-stamp (as a double)
 */
def readServerInfo = { br ->
   double timezoneDifference = 0
   long currentTime = System.currentTimeMillis()

   def serverDateLine = br.readLines().find{it.startsWith(SERVER_DATE_TOKEN) && it.length() > 20}
   if (serverDateLine) {
       // TODO the displayed date-format for perforce appears to be
       //   'yyyy/MM/dd HH:mm:ss Z z' i.e. '2008/08/18 16:02:19 -0400 EDT'
       //   which we could use more directly to get the hr offset and/or the timezone?
       def serverDate = PERFORCE_DATE.parse(serverDateLine.substring(13,32))
       timezoneDifference = getHourDifference(serverDate.getTime(), currentTime)
   }

   // print the results
   println("Time difference between the server and the client is ${timezoneDifference} hours!");
   return timezoneDifference
}

//------------------------------------------------------------------------------
// GET ALL INPUT PARAMETERS and VALIDATE STATE
//------------------------------------------------------------------------------

final def  workDir                   = new File('.').canonicalFile
final def  out                       = System.out
final File resourceHome              = new File(System.getenv()['PLUGIN_HOME'])
final def  props                     = evaluate(new File(resourceHome, 'getAHPProperties.groovy'))

final def client                     = props['source/client']
final boolean isLoggingIn            = props['source/isLoggingIn']?.toBoolean() // TODO should be a source/repo prop?
final def commandPath                = props['source/repo/commandPath'] ?: 'p4'
final def port                       = props['source/repo/port']
final def username                   = props['source/repo/username']
final def password                   = props['source/repo/password'] ?: props['source/repo/passwordScript']
final def charset                    = props['source/repo/charset']
final def commandCharset             = props['source/repo/commandCharset']

def ticketFile                       = props['ticketFile'] != null ? new File(props['ticketFile']) : null;

def clientInfo = null
Double timezoneDifference = null

//------------------------------------------------------------------------------
// Define convenience methods which depend upon input parameters
//------------------------------------------------------------------------------

//create a ProcessBuilder with appropriate ticket file in the environment and the working directory
def pb = new ProcessBuilder(['echo'] as String[]).directory(workDir)
if (ticketFile?.exists()) {
   pb.environment()['P4TICKETS'] = ticketFile.path
}

/**
 * A convenience method for running commands and optionally parsing the stdout of that command.
 * If closure is non-null, the closure will be passed the resultant {@link Process} and is expected to deal with all IO.
 * Otherwise, the process' stdOut and stdErr are forwarded to this scripts 'stdOut' and stdIn is untouched.
 *
 * @param message an optional message to print prior to the commandline
 * @param command the command to be used as a String[]
 * @param closure an optional closure to deal with Process IO
 */
def runCommand = {def message, def command, def closure ->
    pb.command(command as String[])
    println()
    if (message) {
        println(message)
    }
    println("command: ${pb.command().join(' ')}")
    def proc = pb.start()
    if (closure) {
        closure(proc)
    }
    else {
        proc.out.close() // close stdin
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
    proc.waitFor()
    if (proc.exitValue()) {
       throw new Exception("Command failed with exit code: "+proc.exitValue())
    }
}

/**
 * Create a the list of initial arguments needed for perforce commands
 */
def createP4BaseCommand = {
    def command =  [commandPath, '-c', client, '-p', port, '-u', username]
    if (password  && !ticketFile?.exists()) {
        command << '-P' << password
    }
    if (charset) {
        command << '-C' << charset
    }
    if (commandCharset) {
        command << '-Q' << commandCharset
    }
    return command
}

/**
 * Create a the list of initial arguments needed for perforce commands
 */
def createClientlessP4BaseCommand = {
    def command =  [commandPath, '-p', port, '-u', username]
    if (password && !ticketFile?.exists()) {
        command << '-P' << password
    }
    if (charset) {
        command << '-C' << charset
    }
    if (commandCharset) {
        command << '-Q' << commandCharset
    }
    return command
}

/**
 * Run the given script from the plugin resource home directory.  Pass the current properties as the first argument to
 * the script formatted accoring to the java.util.Properties store/load format.
 */
final def runScript = { scriptName ->
    def propsString = new ByteArrayOutputStream()
    def propsArg = new Properties();
    propsArg.putAll(props)
    propsArg.store(propsString, null)
    run(new File(resourceHome, scriptName), [propsString.toString('ISO-8859-1')] as String[])
}

//------------------------------------------------------------------------------
// P4 Login
//------------------------------------------------------------------------------
boolean doLogout = false
if (ticketFile?.exists()) {
    pb.environment()['P4TICKETS'] = ticketFile.path
}
else if (isLoggingIn) {
    ticketFile = File.createTempFile('ah3agent', 'ticket').canonicalFile
    props['ticketFile'] = ticketFile.path
    pb.environment()['P4TICKETS'] = ticketFile.path

    def loginCmd = createP4BaseCommand()
    loginCmd << "login"
    runCommand('P4 Login', loginCmd) { proc ->
        if (password) {
            proc.out << password
        }
        proc.out.close()
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
    doLogout = true
}

//------------------------------------------------------------------------------
// P4 Get Clientspec Info
//------------------------------------------------------------------------------

  def clientInfoCmd = createP4BaseCommand()
  clientInfoCmd << 'client' << '-o'

  runCommand('Geting clientspec...', clientInfoCmd) { proc ->
     proc.consumeProcessErrorStream(System.out)
     proc.in.withReader{ clientInfo = readClientInfo(it) } // read the client info from stdout
  }

  println("Found client root: " + clientInfo.rootPath)
  println("Found client depot paths: ")
  for (depoPath in clientInfo.depoPaths) {
      println("\t${depoPath}")
  }

  //
  // timeDiff
  //

  def serverInfoCmd = createP4BaseCommand()
  serverInfoCmd << 'info'
  runCommand('Getting server info...', serverInfoCmd) { proc ->
      proc.consumeProcessErrorStream(System.out)
      proc.in.withReader{ timezoneDifference = readServerInfo(it) } // read the p4server info from stdout
  }
  props['timezoneDifference'] = timezoneDifference

//------------------------------------------------------------------------------
// AHP Set pertinent info as properties on server
//   1) the timezoneDifference for formatting dates in p4-server timezone
//   2) the rootPath           for set work-dir for job
//   3) depoPathArray          for label step with !mutliLabel attribute
//------------------------------------------------------------------------------

def setPropsCmd = [ahptool, 'setJobProperties', '-']

def propXml = groovy.xml.DOMBuilder.newInstance().'properties'{
    property('name':'timezoneDifference', timezoneDifference)
    property('name':'rootPath', clientInfo.rootPath)
    if (clientInfo.depoPaths) {
        property('name':'depoPathArray', clientInfo.depoPaths.join(','))
    }
}

runCommand('Uploading properties...', setPropsCmd) { proc ->
    proc.consumeProcessOutput(out, out) // forward process stdout and stderr to our stdout
    proc.withWriter{ it << propXml }
}

//------------------------------------------------------------------------------
//P4 Logout
//------------------------------------------------------------------------------
if (doLogout && ticketFile?.exists()) {
   def logoutCmd = createP4BaseCommand()
   logoutCmd << "logout"
   runCommand('P4 Logout', logoutCmd, null)
}