package Perforce
/**
 * 1) P4  login           - if desired
 * 2) P4  get latest change
 * 4) AHP set properties
 * 5) P4  logout          - if login was performed
 * 
 */
 
final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
final def ahptool = isWindows ? 'ahptool.cmd' : 'ahptool'

final def CHANGE_TOKEN ='Change'

final def SERVER_DATE_FORMAT = "yyyy-MM-dd hh:mm:ss.S a z"

public String formatChangelogDateParameter(Long commandDate, Double timeDifference){
    if (!commandDate) {
        return null
    }

    // adjust date for difference between ahp-agent and p4-server
    if (timeDifference) {
        def p4Cal = Calendar.instance
        p4Cal.time = commandDate
        p4Cal.add(Calendar.HOUR_OF_DAY, (int)timeDifference)
        int minDiff = (int)(timeDifference * 60) % 60
        if (minDiff) {
            p4Cal.add(Calendar.MINUTE, minDiff)
        }
        commandDate = p4Cal.time
    }

    def PERFORCE_CHANGELOG_DATE = new java.text.SimpleDateFormat("yyyy/MM/dd:HH:mm:ss")
    return PERFORCE_CHANGELOG_DATE.format(new Date(commandDate))
}

/**
 * read the timezone offset between client and server
 * @param the server info output
 * @return the number of hourse difference between local time and the server time-stamp (as a double)
 */
def readChangeInfo = { br ->

   def changeId
   def changeInfoLine = br.readLines().find{it.startsWith(CHANGE_TOKEN) && it.length() > 7}
   if (changeInfoLine) {
       changeId = changeInfoLine.split(" ")[1]
   }

   // print the results
   println("Latest change found is ${changeId}");
   println("Perforce change line: ${changeInfoLine}");
   return changeId
}

//------------------------------------------------------------------------------
// GET ALL INPUT PARAMETERS and VALIDATE STATE
//------------------------------------------------------------------------------

final def  workDir                   = new File('.').canonicalFile
final def  out                       = System.out
final File resourceHome              = new File(System.getenv()['PLUGIN_HOME'])
final def  props                     = evaluate(new File(resourceHome, 'getAHPProperties.groovy'))

final def client                     = props['source/client']
final boolean isLoggingIn            = props['source/isLoggingIn']?.toBoolean() // TODO should be a source/repo prop?
final def commandPath                = props['source/repo/commandPath'] ?: 'p4'
final def port                       = props['source/repo/port']
final def username                   = props['source/repo/username']
final def password                   = props['source/repo/password'] ?: props['source/repo/passwordScript']
final def charset                    = props['source/repo/charset']
final def commandCharset             = props['source/repo/commandCharset']

def ticketFile                       = props['ticketFile'] != null ? new File(props['ticketFile']) : null;
final def propName                   = props['propertyName']
final Double timeDifference          = props['timezoneDifference']?.toDouble()
final Date workspaceDate             = Date.parse(SERVER_DATE_FORMAT, props['workspace.date'])
final def endDateString              = formatChangelogDateParameter(workspaceDate.getTime(), timeDifference)

def changeId = null

//------------------------------------------------------------------------------
// Define convenience methods which depend upon input parameters
//------------------------------------------------------------------------------

//create a ProcessBuilder with appropriate ticket file in the environment and the working directory
def pb = new ProcessBuilder(['echo'] as String[]).directory(workDir)
if (ticketFile?.exists()) {
   pb.environment()['P4TICKETS'] = ticketFile.absolutePath
}

/**
 * A convenience method for running commands and optionally parsing the stdout of that command.
 * If closure is non-null, the closure will be passed the resultant {@link Process} and is expected to deal with all IO.
 * Otherwise, the process' stdOut and stdErr are forwarded to this scripts 'stdOut' and stdIn is untouched.
 *
 * @param message an optional message to print prior to the commandline
 * @param command the command to be used as a String[]
 * @param closure an optional closure to deal with Process IO
 */
def runCommand = {def message, def command, def closure ->
    pb.command(command as String[])
    println()
    if (message) {
        println(message)
    }
    println("command: ${pb.command().join(' ')}")
    def proc = pb.start()
    if (closure) {
        closure(proc)
    }
    else {
        proc.out.close() // close stdin
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
    proc.waitFor()
    if (proc.exitValue()) {
       throw new Exception("Command failed with exit code: "+proc.exitValue())
    }
}

/**
 * Create a the list of initial arguments needed for perforce commands
 */
def createP4BaseCommand = {
    def command =  [commandPath, '-c', client, '-p', port, '-u', username]
    if (password && !ticketFile?.exists()) {
        command << '-P' << password
    }
    if (charset) {
        command << '-C' << charset
    }
    if (commandCharset) {
        command << '-Q' << commandCharset
    }
    return command
}

//------------------------------------------------------------------------------
// P4 Login
//------------------------------------------------------------------------------
boolean doLogout = false
if (ticketFile?.exists()) {
    pb.environment()['P4TICKETS'] = ticketFile.absolutePath
}
else if (isLoggingIn) {
    ticketFile = File.createTempFile('ah3agent', 'ticket').canonicalFile
    props['ticketFile'] = ticketFile.absolutePath
    pb.environment()['P4TICKETS'] = ticketFile.absolutePath

    def loginCmd = createP4BaseCommand()
    loginCmd << "login"
    runCommand('P4 Login', loginCmd) { proc ->
        if (password) {
            proc.out << password
        }
        proc.out.close()
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
    doLogout = true
}

//------------------------------------------------------------------------------
// P4 Get Clientspec Info
//------------------------------------------------------------------------------

def command = createP4BaseCommand()
command << 'changes' << '-s' << 'submitted'
command << "//${client}/...@${endDateString}"

  runCommand('Getting latest change...', command) { proc ->
      proc.consumeProcessErrorStream(System.out)
      proc.in.withReader{ changeId = readChangeInfo(it) } // read the p4 changes from stdout
  }

//------------------------------------------------------------------------------
// AHP Set pertinent info as properties on server
//   1) the timezoneDifference for formatting dates in p4-server timezone
//   2) the rootPath           for set work-dir for job
//   3) depoPathArray          for label step with !mutliLabel attribute
//------------------------------------------------------------------------------

def setPropsCmd = [ahptool, 'setBuildLifeProperties', '-']

def propXml = groovy.xml.DOMBuilder.newInstance().'properties'{
    property('name':propName, changeId)
}

runCommand('Uploading properties...', setPropsCmd) { proc ->
    proc.consumeProcessOutput(out, out) // forward process stdout and stderr to our stdout
    proc.withWriter{ it << propXml }
}

//------------------------------------------------------------------------------
//P4 Logout
//------------------------------------------------------------------------------
if (doLogout && ticketFile?.exists()) {
   def logoutCmd = createP4BaseCommand()
   logoutCmd << "logout"
   runCommand('P4 Logout', logoutCmd, null)
}