/**
 *
 * Perform:
 *   P4  Login                   - if desired
 *   P4  Get Label Description
 *   P4  Create Label
 *   P4  Apply Label
 *   P4  Logout                  - if login was performed
 *
 */

final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
final def ahptool = isWindows ? 'ahptool.cmd' : 'ahptool'

public final def LINE_SEPARATOR    = System.getProperty('line.separator')
public final def ROOT_TOKEN        = 'Root:'
public final def VIEW_TOKEN        = 'View:'
public final def SERVER_DATE_TOKEN = 'Server date:'
public final def PERFORCE_DATE     = new java.text.SimpleDateFormat("yyyy/MM/dd HH:mm:ss")

//------------------------------------------------------------------------------
// GET ALL INPUT PARAMETERS and VALIDATE STATE
//------------------------------------------------------------------------------

final def  workDir                   = new File('.').canonicalFile
final def  out                       = System.out
final File resourceHome              = new File(System.getenv()['PLUGIN_HOME'])
final def  props                     = evaluate(new File(resourceHome, 'getAHPProperties.groovy'))

final def client                     = props['source/client']
final boolean isLoggingIn            = props['source/isLoggingIn']?.toBoolean() // TODO should be a source/repo prop?
final def commandPath                = props['source/repo/commandPath'] ?: 'p4'
final def port                       = props['source/repo/port']
final def username                   = props['source/repo/username']
final def password                   = props['source/repo/password'] ?: props['source/repo/passwordScript']
final def charset                    = props['source/repo/charset']
final def commandCharset             = props['source/repo/commandCharset']

def ticketFile                       = props['ticketFile'] != null ? new File(props['ticketFile']) : null

String[] depotPathArray      = props['depoPathArray']?.tokenize(',')
boolean multiLabel           = props['source/isLabelingMultipleProjects']
def labelmessage             = props['message']
def label                    = props['label']

if (label =~ /[@#*%\s]/) {
    throw new IllegalStateException('Label name can not contain @#*% or any white-space characters!')
}

if (!depotPathArray) { // null or empty
    throw new IllegalStateException('No depot paths configured')
}

//------------------------------------------------------------------------------
// Define convenience methods which depend upon input parameters
//------------------------------------------------------------------------------

//create a ProcessBuilder with appropriate ticket file in the environment and the working directory
def pb = new ProcessBuilder(['echo'] as String[]).directory(workDir)
if (ticketFile?.exists()) {
   pb.environment()['P4TICKETS'] = ticketFile.absolutePath
}

/**
 * A convenience method for running commands and optionally parsing the stdout of that command.
 * If closure is non-null, the closure will be passed the resultant {@link Process} and is expected to deal with all IO.
 * Otherwise, the process' stdOut and stdErr are forwarded to this scripts 'stdOut' and stdIn is untouched.
 *
 * @param message an optional message to print prior to the commandline
 * @param command the command to be used as a String[]
 * @param closure an optional closure to deal with Process IO
 */
def runCommand = {def message, def command, def closure ->
    pb.command(command as String[])
    println()
    if (message) {
        println(message)
    }
    println("command: ${pb.command().join(' ')}")
    def proc = pb.start()
    if (closure) {
        closure(proc)
    }
    else {
        proc.out.close() // close stdin
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
    proc.waitFor()
    if (proc.exitValue()) {
       throw new Exception("Command failed with exit code: "+proc.exitValue())
    }
}

/**
 * Create a the list of initial arguments needed for perforce commands
 */
def createP4BaseCommand = {
    def command =  [commandPath, '-c', client, '-p', port, '-u', username]
    if (password && !ticketFile?.exists()) {
        command << '-P' << password
    }
    if (charset) {
        command << '-C' << charset
    }
    if (commandCharset) {
        command << '-Q' << commandCharset
    }
    return command
}

/**
 * Run the given script from the plugin resource home directory.  Pass the current properties as the first argument to
 * the script formatted accoring to the java.util.Properties store/load format.
 */
final def runScript = { scriptName ->
    def propsString = new ByteArrayOutputStream()
    def propsArg = new Properties();
    propsArg.putAll(props)
    propsArg.store(propsString, null)
    run(new File(resourceHome, scriptName), [propsString.toString('ISO-8859-1')] as String[])
}

//------------------------------------------------------------------------------
// P4 Login
//------------------------------------------------------------------------------
boolean doLogout = false
if (ticketFile?.exists()) {
    pb.environment()['P4TICKETS'] = ticketFile.absolutePath
}
else if (isLoggingIn) {
    ticketFile = File.createTempFile('ah3agent', 'ticket').canonicalFile
    props['ticketFile'] = ticketFile.absolutePath
    pb.environment()['P4TICKETS'] = ticketFile.absolutePath

    def loginCmd = createP4BaseCommand()
    loginCmd << "login"
    runCommand('P4 Login', loginCmd) { proc ->
        if (password) {
            proc.out << password
        }
        proc.out.close()
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
    doLogout = true
}

//------------------------------------------------------------------------------
// P4 get description of label (generate a label-specification template text)
//------------------------------------------------------------------------------
def labelDescription = null

def getLabelDescCmd = createP4BaseCommand()
getLabelDescCmd << 'label' << '-o' << label
runCommand('Getting Label Description', getLabelDescCmd) { proc ->
    proc.out.close() // close stdin
    proc.consumeProcessErrorStream(System.out)
    labelDescription = proc.text // read the label description
}

//------------------------------------------------------------------------------
// edit the label for description and viewpath
//------------------------------------------------------------------------------
def modLabelDescription = new StringBuilder()
labelDescription.eachLine{line ->
    modLabelDescription << line + LINE_SEPARATOR
    if (labelmessage && line.startsWith('Description: ')) {
        modLabelDescription << "\t${labelmessage}" + LINE_SEPARATOR
    }
    else if (!multiLabel && line.startsWith('View: ')) {
        for (path in depotPathArray) {
            modLabelDescription << "\t${path}" + LINE_SEPARATOR
        }
    }
}
labelDescription = modLabelDescription.toString()

//------------------------------------------------------------------------------
// P4  save the label
//------------------------------------------------------------------------------

def saveLabelCmd = createP4BaseCommand()
saveLabelCmd << 'label' << '-i'
runCommand('Save Label', saveLabelCmd) { proc ->
    proc.consumeProcessOutput(out, out)
    proc.withWriter{ it << labelDescription } // feed in the modified label definition to the process (and close/flush stream afterwards)
}

//------------------------------------------------------------------------------
// P4 apply the label
//------------------------------------------------------------------------------

def applyLableCmd = createP4BaseCommand()
applyLableCmd << 'labelsync'
if (multiLabel) {
    applyLableCmd << '-a'
}
applyLableCmd << '-l' << label
runCommand('Apply Label', applyLableCmd, null)

//------------------------------------------------------------------------------
//P4 Logout
//------------------------------------------------------------------------------
if (doLogout && ticketFile?.exists()) {
   def logoutCmd = createP4BaseCommand()
   logoutCmd << "logout"
   runCommand('P4 Logout', logoutCmd, null)
}