/*
 * Does not use the current workdir.
 * Perform a changelog and upload the latest change date from within the desired interval to ahp
 *
 * 1) P4  login             - if desired
 * 2) P4  create clientspec - if using template, also edit spec if using template with custom work-dir
 * 3) P4  changes           - (also a summary for each change found to get files and more accurate timestamps)
 * 4) AHP setprop latest change date
 * 5) P4  delete clientspec - if using template
 * 6) P4  logout            - if login was performed
 *
 */

final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
final def ahptool = isWindows ? 'ahptool.cmd' : 'ahptool'

final def PERFORCE_CHANGELOG_OUT_DATE = new java.text.SimpleDateFormat("yyyy/MM/dd HH:mm:ss")
final def CHANGE_SUMMARY_PATTERN = ~'(?m)^Change \\d+ by (\\S+)@\\S+ on (\\d+/\\d+/\\d+ \\d+:\\d+:\\d+)'
final def ROOT_TOKEN        ='Root:'
final def VIEW_TOKEN        ='View:'
final def SERVER_DATE_TOKEN ='Server date:'
final def PERFORCE_DATE     = new java.text.SimpleDateFormat("yyyy/MM/dd HH:mm:ss")

/**
 * @param base the base time in milliseconds since Epoc
 * @param target the target time in milliseconds since Epoc
 * @return the number of hours between the two times (rounded to nearest half-hour)
 */
public double getHourDifference(long base, long target) {
    double difference = (double)(base - target) / (1000*60*60) // difference in hours
    return Math.round(difference / 0.5) * 0.5                  // round to nearest 1/2 hour
}

/**
 *  read stdout of get-client-info afor the workspacepath and depoPathArray
 *  @param br the buffered reader of the p4 command output
 *  @return a map containing 'rootPath':String and 'depoPaths':String[]
 */
final def readClientInfo = {def br ->
    def result = [rootPath: null, depoPaths: []]

    String line = null
    while( (line = br.readLine()) != null )  {
        if (line.startsWith(ROOT_TOKEN)) {
            result.rootPath = line.substring(ROOT_TOKEN.length()).trim()
        }
        else if (line.startsWith(VIEW_TOKEN)) {
            while (line = br.readLine()?.trim()) { // read until we hit a blank line (or null)
                // match first token '//....' where ... has no white-space OR '"//...."' where ... has no quote
                def m = ( line =~ '((?:"//[^"]+)|(?://\\S+))' )
                m.find()
                result.depoPaths << m.group(1)
            }
        }
    }

    return result
}
/**
 * read the timezone offset between client and server
 */
final def readServerInfo = { br ->
    double timezoneDifference = 0
    long currentTime = System.currentTimeMillis()

    def serverDateLine = br.readLines().find{it.startsWith(SERVER_DATE_TOKEN) && it.length() > 20}
    if (serverDateLine) {
       // TODO the displayed date-format for perforce appears to be
       //   'yyyy/MM/dd HH:mm:ss Z z' i.e. '2008/08/18 16:02:19 -0400 EDT'
       //   which we could use more directly to get the hr offset and/or the timezone?
       def serverDate = PERFORCE_DATE.parse(serverDateLine.substring(13,32))
       timezoneDifference = getHourDifference(serverDate.getTime(), currentTime)
    }

    // print the results
    println("Time difference between the server and the client is ${timezoneDifference} hours!");
    return timezoneDifference
}

/**
 *
 */
public String formatChangelogDateParameter(Long commandDate, Double timeDifference){
    if (!commandDate) {
        return null
    }

    def PERFORCE_CHANGELOG_DATE = new java.text.SimpleDateFormat("yyyy/MM/dd:HH:mm:ss")

    // adjust date for difference between ahp-agent and p4-server
    if (timeDifference) {
        def timeDate = new Date(commandDate)
        def p4Cal = Calendar.instance
        p4Cal.setTime(timeDate)
        p4Cal.add(Calendar.HOUR_OF_DAY, (int)timeDifference)
        int minDiff = (int)(timeDifference * 60) % 60
        if (minDiff) {
            p4Cal.add(Calendar.MINUTE, minDiff)
        }
        commandDate = p4Cal.getTimeInMillis()
    }

    return PERFORCE_CHANGELOG_DATE.format(new Date(commandDate))
}



//------------------------------------------------------------------------------
// GET ALL INPUT PARAMETERS and VALIDATE STATE
//------------------------------------------------------------------------------

final def  workDir                   = new File('.').canonicalFile
final def  out                       = System.out
final File resourceHome              = new File(System.getenv()['PLUGIN_HOME'])
final def  props                     = evaluate(new File(resourceHome, 'getAHPProperties.groovy'))

final def client                     = props['source/client']
//final def workspacePath              = null // TODO
final boolean isLoggingIn            = props['source/isLoggingIn']?.toBoolean() // TODO should be a source/repo prop?
final def commandPath                = props['source/repo/commandPath'] ?: 'p4'
final def port                       = props['source/repo/port']
final def username                   = props['source/repo/username']
final def password                   = props['source/repo/password'] ?: props['source/repo/passwordScript']
final def charset                    = props['source/repo/charset']
final def commandCharset             = props['source/repo/commandCharset']

final boolean isUsingTemplate        = props['source/clientSpecPolicy'] == 'template'
final boolean createNewClientSpec    = props['source/clientSpecPolicy'] == 'literal'
final def templateName               = props['source/clientspecTemplateName'] // only if 'source/isUsingTemplate' is true
final def clientspec                 = props['source/clientspec']

//job properties (these values currently only come from getclientinfo and p4login respectively)
final Double timeDifference = props['timezoneDifference']?.toDouble()
final def File ticketFile        = props['ticketFile']?new File(props['ticketFile']):null

// start and end dates
final Long startDate       = props['startDate']?.toLong()
final Long endDate         = props['endDate']?.toLong()

long latestChangeDate = 0

//create a ProcessBuilder with appropriate ticket file in the environment and the working directory
def pb = new ProcessBuilder(['echo'] as String[]).directory(workDir)
if (ticketFile?.exists()) {
   pb.environment()['P4TICKETS'] = ticketFile.absolutePath
}

//------------------------------------------------------------------------------
// Define convenience methods which depend upon input parameters
//------------------------------------------------------------------------------

/**
 * A convenience method for running commands and optionally parsing the stdout of that command.
 * If closure is non-null, the closure will be passed the resultant {@link Process} and is expected to deal with all IO.
 * Otherwise, the process' stdOut and stdErr are forwarded to this scripts 'stdOut' and stdIn is untouched.
 *
 * @param message an optional message to print prior to the commandline
 * @param command the command to be used as a String[]
 * @param closure an optional closure to deal with Process IO
 */
final def runCommand = {def message, def command, def closure ->
    pb.command(command as String[])
    println()
    if (message) {
        println(message)
    }
    println("command: ${pb.command().join(' ')}")
    def proc = pb.start()
    if (closure) {
        closure(proc)
    }
    else {
        proc.out.close() // close stdin
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
    proc.waitFor()
    if (proc.exitValue()) {
       throw new Exception("Command failed with exit code: "+proc.exitValue())
    }
}

/**
 * Create a the list of initial arguments needed for perforce commands
 */
final def createP4BaseCommand = {
    def command =  [commandPath, '-c', client, '-p', port, '-u', username]
    if (password  && !ticketFile?.exists()) {
        command << '-P' << password
    }
    if (charset) {
        command << '-C' << charset
    }
    if (commandCharset) {
        command << '-Q' << commandCharset
    }
    return command
}

/**
 * Run the given script from the plugin resource home directory.  Pass the current properties as the first argument to
 * the script formatted accoring to the java.util.Properties store/load format.
 */
final def runScript = { scriptName ->
    def propsString = new ByteArrayOutputStream()
    def propsArg = new Properties();
    propsArg.putAll(props)
    propsArg.store(propsString, null)
    run(new File(resourceHome, scriptName), [propsString.toString('ISO-8859-1')] as String[])
}

//------------------------------------------------------------------------------
// P4 Login
//------------------------------------------------------------------------------
boolean doLogout = false
if (ticketFile?.exists()) {
    pb.environment()['P4TICKETS'] = ticketFile.absolutePath
}
else if (isLoggingIn) {
    ticketFile = File.createTempFile('ah3agent', 'ticket').canonicalFile
    props['ticketFile'] = ticketFile.absolutePath
    pb.environment()['P4TICKETS'] = ticketFile.absolutePath

    def loginCmd = createP4BaseCommand()
    loginCmd << "login"
    runCommand('P4 Login', loginCmd) { proc ->
        if (password) {
            proc.out << password
        }
        proc.out.close()
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
    doLogout = true
}

//------------------------------------------------------------------------------
// P4 Create Clientspec (if needed)
//------------------------------------------------------------------------------

runScript('p4_createclientspec.groovy')

//------------------------------------------------------------------------------
// P4 Get Clientspec Info
//------------------------------------------------------------------------------

  //
  // client info
  //

  def clientInfoCmd = createP4BaseCommand()
  clientInfoCmd << 'client' << '-o'

  runCommand('Geting clientspec...', clientInfoCmd) { proc ->
     proc.consumeProcessErrorStream(System.out)
     proc.in.withReader{ clientInfo = readClientInfo(it) } // read the client info from stdout
  }

  println("Found client root: " + clientInfo.rootPath)
  println("Found client depot paths: ")
  for (depoPath in clientInfo.depoPaths) {
      println("\t${depoPath}")
  }

  //
  // timeDiff
  //

  def serverInfoCmd = createP4BaseCommand()
  serverInfoCmd << 'info'
  runCommand('Getting server info...', serverInfoCmd) { proc ->
      proc.consumeProcessErrorStream(System.out)
      proc.in.withReader{ timezoneDifference = readServerInfo(it) } // read the p4server info from stdout
  }

//------------------------------------------------------------------------------
// Get Change Set Ids
//------------------------------------------------------------------------------
final def startDateString = formatChangelogDateParameter(startDate, timezoneDifference) ?: null
final def endDateString   = formatChangelogDateParameter(endDate,   timezoneDifference) ?: 'now'

final def revisionSpecifier
if (startDateString) {
    revisionSpecifier = "//${client}/...@${startDateString},${endDateString}"
}
else if (endDateString) { // isn't this always true?
    revisionSpecifier = "//${client}/...@${endDateString}"
}
else {
    revisionSpecifier = "//${client}/..."
}

def changes = null

def changesCmd = createP4BaseCommand()
changesCmd << 'changes' << '-s' << 'submitted' << revisionSpecifier
runCommand('Getting Changes', changesCmd) { proc ->
    proc.consumeProcessErrorStream(System.out)
    changes = proc.text // get the stdout of the process
    proc.in.close() // close the stdin of the process
}
println("found ${changes.readLines().size()} change-sets")

//
// Get dates of changes for changesets with files/author
//
changes.eachLine{ line ->
    def matcher = ( line =~ /(?m)^Change (\d+)/ ) // Change X on yyyy/MM/dd by user@client 'comment'
    if (matcher) { // find pattern
        int changeId = matcher.group(1).toInteger()
        def summary = ''

        def descCommand = createP4BaseCommand()
        descCommand << 'describe' << '-s' << changeId
        runCommand(null, descCommand) { proc ->
            proc.consumeProcessErrorStream(System.out)
            summary = proc.text // get the stdout of the process
            proc.in.close() // close the stdin of the process
        }

        // first line is "Change X by user@client on yyyy/MM/dd HH:mm:ss"
        def summaryMatcher = CHANGE_SUMMARY_PATTERN.matcher(summary)
        if (summaryMatcher) {
            boolean hasAllowedFile = true   // TODO check for one or more files within allowed files
            boolean hasAllowedAuthor = true // TODO check user is in allowed users

            def user  = summaryMatcher.group(1)
            def date = PERFORCE_CHANGELOG_OUT_DATE.parse(summaryMatcher.group(2))
            // TODO deal with 'unknown' date

            // Evenatially there is a line "Affected files ..."
            //  followed by lines of "... //depot/file#3 edit"
            summary.eachMatch(/(?m)^\.\.\. (.+)#\d+ (\S+)/){entry, file, operation ->
                hasAllowedFile = hasAllowedFile || true // do an early-excape ||= type operation for detecting an allowed file
            }

            if (hasAllowedAuthor && hasAllowedFile) {
                latestChangeDate = Math.max(latestChangeDate, date.time)
            }
        }
    }
}

//------------------------------------------------------------------------------
//  UPLOAD LATEST CHANGE DATE
// ------------------------------------------------------------------------------

if (!latestChangeDate) {
    println('No changes detected')
}
else {
    def uploadChangeDateCmd = [ahptool, 'setStepProperties', '-']

    def propXml = groovy.xml.DOMBuilder.newInstance().'properties'{
        property('name':'latest-change', latestChangeDate)
    }

    runCommand('Uploading time of latest change...', uploadChangeDateCmd) { proc ->
        proc.consumeProcessOutput(out, out) // forward process stdout and stderr to our stdout
        proc.withWriter{ it << propXml }
    }
}

//------------------------------------------------------------------------------
// Delete Clientspec (If needed)
//------------------------------------------------------------------------------
runScript('p4_deleteclientspec.groovy')

//------------------------------------------------------------------------------
//P4 Logout
//------------------------------------------------------------------------------
if (doLogout && ticketFile?.exists()) {
   def logoutCmd = createP4BaseCommand()
   logoutCmd << "logout"
   runCommand('P4 Logout', logoutCmd, null)
}