/**
 * Handle all substeps involed in performing a p4 sync
 *
 * 1) P4  login             - if desired
 * 2) P4  create clientspec - if using template, also edit spec if using template with custom work-dir
 * 3) P4  getclientinfo     - to get root, depotPaths, and timezone offset
 * 4) AHP set workdir
 * 5) P4  sync
 * 6) P4  logout            - if login was performed
 *
 */

final def PERFORCE_DATE           = new java.text.SimpleDateFormat("yyyy/MM/dd HH:mm:ss")

//------------------------------------------------------------------------------
// GET ALL INPUT PARAMETERS and VALIDATE STATE
//------------------------------------------------------------------------------

final def  workDir                   = new File('.').canonicalFile
final def  out                       = System.out
final File resourceHome              = new File(System.getenv()['PLUGIN_HOME'])
final def  props                     = evaluate(new File(resourceHome, 'getAHPProperties.groovy'))

final def client                     = props['source/client']
final boolean isLoggingIn            = props['source/isLoggingIn']?.toBoolean() // TODO should be a source/repo prop?
final def commandPath                = props['source/repo/commandPath'] ?: 'p4'
final def port                       = props['source/repo/port']
final def username                   = props['source/repo/username']
final def password                   = props['source/repo/password'] ?: props['source/repo/passwordScript']
final def charset                    = props['source/repo/charset']
final def commandCharset             = props['source/repo/commandCharset']

final boolean forceSync              = props['source/isForceSync'] ? props['source/isForceSync'].toBoolean() : false
final def label                      = props['source/label']
final boolean preserveUnlabeledFiles = props['source/preserveFilesWithoutLabel']?.toBoolean()
final boolean doNotUpdateHaveList    = props['source/doNotUpdateHaveList']?.toBoolean()

def ticketFile                       = props['ticketFile'] != null ? new File(props['ticketFile']) : null

def date                             = props['date']?.toLong()
if (date) {
    date = new Date(date)
}

// from Login/GetClientInfo step (set as job property if present)
Double timezoneDifference      = props['timezoneDifference']?.toDouble()


//create a ProcessBuilder with appropriate ticket file in the environment and the working directory
def pb = new ProcessBuilder(['echo'] as String[]).directory(workDir)
if (ticketFile?.exists()) {
   pb.environment()['P4TICKETS'] = ticketFile.absolutePath
}

//------------------------------------------------------------------------------
// Define convenience methods which depend upon input parameters
//------------------------------------------------------------------------------

/**
 * A convenience method for running commands and optionally parsing the stdout of that command.
 * If closure is non-null, the closure will be passed the resultant {@link Process} and is expected to deal with all IO.
 * Otherwise, the process' stdOut and stdErr are forwarded to this scripts 'stdOut' and stdIn is untouched.
 *
 * @param message an optional message to print prior to the commandline
 * @param command the command to be used as a String[]
 * @param closure an optional closure to deal with Process IO
 */
def runCommand = {def message, def command, def closure ->
    pb.command(command as String[])
    println()
    if (message) {
        println(message)
    }
    println("command: ${pb.command().join(' ')}")
    def proc = pb.start()
    if (closure) {
        closure(proc)
    }
    else {
        proc.out.close() // close stdin
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
    proc.waitFor()
    if (proc.exitValue()) {
       throw new Exception("Command failed with exit code: "+proc.exitValue())
    }
}

/**
 * Create a the list of initial arguments needed for perforce commands
 */
def createP4BaseCommand = {
    def command =  [commandPath, '-c', client, '-p', port, '-u', username]
    if (password && !ticketFile?.exists()) {
        command << '-P' << password
    }
    if (charset) {
        command << '-C' << charset
    }
    if (commandCharset) {
        command << '-Q' << commandCharset
    }
    return command
}

/**
 * Run the given script from the plugin resource home directory.  Pass the current properties as the first argument to
 * the script formatted accoring to the java.util.Properties store/load format.
 */
final def runScript = { scriptName ->
    def propsString = new ByteArrayOutputStream()
    def propsArg = new Properties();
    propsArg.putAll(props)
    propsArg.store(propsString, null)
    run(new File(resourceHome, scriptName), [propsString.toString('ISO-8859-1')] as String[])
}

//------------------------------------------------------------------------------
// P4 Login
//------------------------------------------------------------------------------
boolean doLogout = false
if (ticketFile?.exists()) {
    pb.environment()['P4TICKETS'] = ticketFile.absolutePath
}
else if (isLoggingIn) {
    ticketFile = File.createTempFile('ah3agent', 'ticket').canonicalFile
    props['ticketFile'] = ticketFile.absolutePath
    pb.environment()['P4TICKETS'] = ticketFile.absolutePath

    def loginCmd = createP4BaseCommand()
    loginCmd << "login"
    runCommand('P4 Login', loginCmd) { proc ->
        if (password) {
            proc.out << password
        }
        proc.out.close()
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
    }
    doLogout = true
}

//------------------------------------------------------------------------------
// P4 Sync
//------------------------------------------------------------------------------

def syncCmd = createP4BaseCommand()
syncCmd << 'sync'
if (forceSync) {
    syncCmd << '-f'
}
else if (doNotUpdateHaveList) {
    syncCmd << '-p'
}

if (label) {
    if (preserveUnlabeledFiles) {
        syncCmd << "@$label,@$label"
    }
    else {
        syncCmd << "@$label"
    }
}
else if (date != null) {
    // take ahp3-server date and format it using the p4 server timezone
    def commandDate = date

    // adjust date for difference between ahp-agent and p4-server
    if (timezoneDifference) {
        def p4Cal = Calendar.instance
        p4Cal.time = commandDate
        p4Cal.add(Calendar.HOUR_OF_DAY, (int)timezoneDifference)
        int minDiff = (int)(timezoneDifference * 60) % 60
        if (minDiff) {
            p4Cal.add(Calendar.MINUTE, minDiff)
        }
        commandDate = p4Cal.time
    }

    syncCmd << "@${PERFORCE_DATE.format(commandDate)}"
}

runCommand("Performing P4 sync...", syncCmd, null)

//------------------------------------------------------------------------------
//P4 Logout
//------------------------------------------------------------------------------
if (doLogout && ticketFile?.exists()) {
   def logoutCmd = createP4BaseCommand()
   logoutCmd << "logout"
   runCommand('P4 Logout', logoutCmd, null)
}